/*
 * $Id: Request.java 278 2005-11-21 22:34:14Z davidb $
 * 
 * Copyright (c) 2004 USC Project SRO. All rights reserved.
 * Copyright (c) 2005 VeriSign, Inc. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;

import org.xbill.DNS.*;

/**
 * This class represents a DNS request, which encompasses the actual question
 * asked, as well as any ancillary data needed to send and process the
 * request.
 * 
 * @author davidb
 * @version $Revision: 278 $
 */
public class Request implements Cloneable
{
  private Name   mQName;
  private int    mQType;
  private int    mQClass;
  private BitSet mFlags;
  private int    mOpcode;
  private int    mID;
  private int    mEDNSVersion;
  private int    mBufsize;
  private int    mExtendedRcode;
  private int    mExtendedFlags;
  private List   mEDNSOptions;

  // send state.
  private Host   mTarget;
  private int    mWakeupTime;

  /**
   * Create an empty request.
   */
  public Request()
  {
    mFlags = new BitSet();
    mEDNSVersion = -1;
    mOpcode = Opcode.QUERY; // by default.
  }

  /**
   * Create a request for the given query trinity.
   * 
   * @param qname The name to query for.
   * @param qtype The type to query for.
   * @param qclass The class to query for.
   */
  public Request(Name qname, int qtype, int qclass)
  {
    this();
    setQName(qname);
    mQType = qtype;
    mQClass = qclass;
  }

  /**
   * Create a request from the given query message
   * 
   * @param m a query message (QR should be zero).
   */
  public Request(Message m)
  {
    this();
    Record q = m.getQuestion();
    setQName(q.getName());
    mQType = q.getType();
    mQClass = q.getDClass();
    setHeader(m.getHeader());
    setOPT(m.getOPT());
  }

  /**
   * Clone a requst. The clone will share a reference to the query context
   * parameters.
   */
  public Object clone()
  {
    try
    {
      Request r = (Request) super.clone();
      r.mFlags = (BitSet) mFlags.clone();
      if (mEDNSOptions != null)
      {
        r.mEDNSOptions = new ArrayList();
        r.mEDNSOptions.addAll(mEDNSOptions);
      }
      return r;
    }
    catch (CloneNotSupportedException e)
    {}

    return null;
  }

  public static Request newQuery(Request request, Name qname)
  {
    Request r = (Request) request.clone();
    r.setQName(qname);

    return r;
  }

  /**
   * @return the qname for this request.
   */
  public Name getQName()
  {
    return mQName;
  }

  /**
   * Set the qname for this request.
   * 
   * @param name the name to set.
   */
  public void setQName(Name name)
  {
    mQName = name;
  }

  /**
   * @return the query type for this request.
   */
  public int getQType()
  {
    return mQType;
  }

  /**
   * @return The query class for this request.
   */
  public int getQClass()
  {
    return mQClass;
  }

  /**
   * @return
   */
  public Record getQuestion()
  {
    return Record.newRecord(mQName, mQType, mQClass);
  }

  public boolean getFlag(byte flag)
  {
    return mFlags.get(flag);
  }

  public void setFlag(byte flag)
  {
    mFlags.set(flag);
  }

  public void unsetFlag(byte flag)
  {
    mFlags.clear(flag);
  }

  public void setFlag(byte flag, boolean value)
  {
    mFlags.set(flag, value);
  }

  public int getID()
  {
    return mID;
  }

  public void setID(int id)
  {
    mID = id;
  }

  public boolean hasOPT()
  {
    return mEDNSVersion >= 0;
  }

  public int getBufsize()
  {
    return mBufsize;
  }

  public void setBufsize(int bufsize)
  {
    mBufsize = bufsize;
  }

  public List getEDNSOptions()
  {
    return mEDNSOptions;
  }

  public void setEDNSOptions(List options)
  {
    mEDNSOptions = options;
  }

  public int getEDNSVersion()
  {
    return mEDNSVersion;
  }

  public void setEDNSVersion(int version)
  {
    mEDNSVersion = version;
  }

  public int getExtendedFlags()
  {
    return mExtendedFlags;
  }

  public void setExtendedFlags(int extendedFlags)
  {
    mExtendedFlags = extendedFlags;
  }

  public int getExtendedRcode()
  {
    return mExtendedRcode;
  }

  public void setExtendedRcode(int extendedRcode)
  {
    mExtendedRcode = extendedRcode;
  }

  public void setHeader(Header h)
  {
    for (int i = 0; i < 16; i++)
    {
      if (Flags.isFlag(i)) mFlags.set(i, h.getFlag(i));
    }

    mOpcode = h.getOpcode();
    mID = h.getID();
  }

  /**
   * @return The OPT record for this request, or null if there isn't one.
   */
  public OPTRecord getOPT()
  {
    if (!hasOPT()) return null;
    OPTRecord opt = new OPTRecord(mBufsize, mExtendedRcode, mEDNSVersion,
        mExtendedFlags, mEDNSOptions);
    return opt;
  }

  /**
   * Set the OPT record for this request.
   * 
   * @param opt
   */
  public void setOPT(OPTRecord opt)
  {
    if (opt == null)
    {
      setEDNSVersion(-1);
      return;
    }
    setEDNSVersion(opt.getVersion());
    setBufsize(opt.getPayloadSize());
    setExtendedRcode(opt.getExtendedRcode());
    setExtendedFlags(opt.getFlags());
    setEDNSOptions(opt.getOptions());
  }

  /**
   * Note that this returns a new Message object from the request, so
   * modifying this message does not change the request in any way.
   * 
   * @return Returns a message constructed from this request.
   */
  public Message getMessage()
  {
    Record queryRecord = getQuestion();
    Message m = Message.newQuery(queryRecord);

    // copy the header info into the existing message
    Header h = m.getHeader();
    h.setOpcode(mOpcode);
    h.setID(mID);
    for (int i = 0; i < mFlags.length(); i++)
    {
      if (Flags.isFlag(i)) h.setFlag(i, mFlags.get(i));
    }

    if (hasOPT())
    {
      m.addRecord(getOPT(), Section.ADDITIONAL);
    }

    return m;
  }

  public boolean equivalent(Request r)
  {
    if (r.getQType() != getQType()) return false;
    if (r.getQClass() != getQClass()) return false;
    if (!r.getQName().equals(getQName())) return false;

    return true;
  }

  public boolean equals(Object o)
  {
    Request r = (Request) o;
    if (!equivalent(r)) return false;
    if (r.getID() != getID()) return false;

    return true;
  }

  public int hashCode()
  {
    // FIXME: probably need to do something more clever for the hashcode here.
    return mQName.hashCode() + mQType + mQClass + mID;
  }

  public String toString()
  {
    return getQName() + "/" + Type.string(getQType()) + "/"
        + DClass.string(getQClass());
  }

  /**
   * @return Returns the target.
   */
  public Host getTarget()
  {
    return mTarget;
  }

  /**
   * @param target The target to set.
   */
  public void setTarget(Host target)
  {
    mTarget = target;
  }

  /**
   * @return Returns the wakeupTime.
   */
  public int getWakeupTime()
  {
    return mWakeupTime;
  }

  /**
   * @param wakeupTime The wakeupTime to set.
   */
  public void setWakeupTime(int wakeupTime)
  {
    mWakeupTime = wakeupTime;
  }

}