/*
 * $Id: DigTest.java 238 2005-08-15 22:26:43Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound.cl;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.PropertyConfigurator;
import org.xbill.DNS.DClass;
import org.xbill.DNS.Flags;
import org.xbill.DNS.Message;
import org.xbill.DNS.Name;
import org.xbill.DNS.Rcode;
import org.xbill.DNS.TextParseException;
import org.xbill.DNS.Type;

import se.rfc.unbound.DNSEvent;
import se.rfc.unbound.Request;
import se.rfc.unbound.Resolver;
import se.rfc.unbound.iter.IterativeResolver;
import se.rfc.unbound.nsclient.NameserverClient;
import se.rfc.unbound.stub.StubResolver;
import se.rfc.unbound.validator.ValidatingResolver;

public class DigTest
{
  private static class CLIState
  {
    public Options    opts;
    public Properties config;
    public boolean    debug;
    public boolean    trafficDebug;
    public Name       qname;
    public int        qtype;
    public int        qclass;
    public int        qport;
    public String     qtarget;
    public boolean    useDNSSEC;
    public boolean    noChecking;

    public boolean    useIterativeResolver = false;

    // etc...

    public CLIState()
    {
      opts = new Options();

      // boolean options
      opts.addOption("h", "help", false, "Print this message.");
      opts.addOption("v", "verbose", false, "Increase verbosity.");
      opts.addOption("T", "traffic", false, "Log DNS traffic");

      OptionBuilder.hasArg();
      OptionBuilder.withLongOpt("config");
      OptionBuilder.withArgName("config_file");
      OptionBuilder.withDescription("overriding configuration file");
      opts.addOption(OptionBuilder.create('c'));

      OptionBuilder.hasArg();
      OptionBuilder.withLongOpt("port");
      OptionBuilder.withArgName("port");
      OptionBuilder.withDescription("port to use when querying (default 53)");
      opts.addOption(OptionBuilder.create('p'));

      OptionBuilder.withLongOpt("use-dnssec");
      OptionBuilder.withDescription("enable DNSSEC");
      opts.addOption(OptionBuilder.create('D'));

      OptionBuilder.withLongOpt("disable-checking");
      OptionBuilder.withDescription("Do not validate DNSSEC responses.");
      opts.addOption(OptionBuilder.create('C'));

      OptionBuilder.withLongOpt("use-iterative");
      OptionBuilder
          .withDescription("Use the Iterative Resolver module directly.");
      opts.addOption(OptionBuilder.create('I'));
    }

    public void parseCommandLine(String[] args) throws ParseException
    {
      CommandLineParser cli_parser = new PosixParser();
      CommandLine cli = cli_parser.parse(opts, args);

      if (cli.hasOption('h')) usage();
      if (cli.hasOption('v')) debug = true;
      if (cli.hasOption('T')) trafficDebug = true;
      if (cli.hasOption('D')) useDNSSEC = true;
      if (cli.hasOption('C')) noChecking = true;
      if (cli.hasOption('I')) useIterativeResolver = true;

      try
      {
        if (cli.getOptionValue('c') != null)
        {
          config = readConfiguration(cli.getOptionValue('c'));
        }
        else
        {
          config = defaultConfiguration();
        }
      }
      catch (IOException e)
      {
        throw new ParseException(e.getMessage());
      }

      setupLogging();

      if (debug)
      {
        Logger rootLogger = Logger.getRootLogger();
        rootLogger.setLevel(Level.DEBUG);
        rootLogger.debug("Debug logging on.");
      }
      Logger l = Logger.getLogger("traffic_log");
      if (trafficDebug)
      {
        l.setLevel(Level.DEBUG);
        l.debug("logging traffic.");
      }
      else
      {
        l.setLevel(Level.WARN);
        l.warn("Not logging traffic.");
      }

      if (cli.getOptionValue('p') != null)
      {
        try
        {
          qport = Integer.parseInt(cli.getOptionValue('p'));
        }
        catch (NumberFormatException e)
        {
          System.err.println("error: port number invalid");
        }
      }

      // Parse the rest of the command line, which looks like:
      // [@target] <qname> [qtype] [qclass]
      String[] a = cli.getArgs();
      if (a.length < 1)
      {
        System.err.println("error: a query name must be specified");
        usage();
      }

      qtype = -1;
      qclass = -1;
      for (int i = 0; i < a.length; i++)
      {
        // Check for alternate target.
        if (a[i].startsWith("@"))
        {
          qtarget = a[i].substring(1);
          continue;
        }

        // The first eligible string is the qname.
        if (qname == null)
        {
          try
          {
            qname = Name.fromString(a[i]);
          }
          catch (TextParseException e)
          {
            throw new ParseException(e.getMessage());
          }
          continue;
        }

        if (qtype < 0 && Type.value(a[i]) > 0)
        {
          qtype = Type.value(a[i]);
          continue;
        }
        if ((qclass < 0) && (DClass.value(a[i]) > 0))
        {
          qclass = DClass.value(a[i]);
        }
      }

      if (qtype < 0) qtype = Type.A;
      if (qclass < 0) qclass = DClass.IN;
    }

    /**
     * Instantiate a configuration from the given Java properties file.
     * 
     * @param filename A valid Java properties file.
     * @return The parsed configuration.
     */
    public static Properties readConfiguration(String filename)
        throws IOException
    {
      Properties config = new Properties();
      InputStream file = new FileInputStream(filename);
      config.load(file);
      file.close();

      return config;
    }

    public static Properties defaultConfiguration()
    {
      Properties config = new Properties();
      // Set any default properties here...
      return config;
    }

    public boolean hasLog4jConfig(Properties cf)
    {
      for (Iterator i = cf.keySet().iterator(); i.hasNext();)
      {
        String key = (String) i.next();
        if (key.startsWith("log4j."))
        {
          return true;
        }
      }
      return false;
    }

    public void setupLogging()
    {
      String s = config.getProperty("log4j.config");
      if (s != null)
      {
        PropertyConfigurator.configure(s);
      }
      else if (hasLog4jConfig(config))
      {
        PropertyConfigurator.configure(config);
      }
      else
      {
        // use a shorter layout than the TTCCC layout
        Logger root = Logger.getRootLogger();
        root.addAppender(new ConsoleAppender(
            new PatternLayout("%d [%p] %m\n")));
      }
    }

    /**
     * This is a basic usage statement that generates the actual usage from
     * the command line options object, and then quits.
     * 
     * @param opts the command line options, generally created by the
     *          setupCLI() method.
     */
    private void usage()
    {
      HelpFormatter f = new HelpFormatter();

      PrintWriter out = new PrintWriter(System.err);

      // print our own usage statement:
      f.printHelp(out,
          75,
          "digTest.sh [..options..] qname [qtype] [qclass]",
          null,
          opts,
          HelpFormatter.DEFAULT_LEFT_PAD,
          HelpFormatter.DEFAULT_DESC_PAD,
          null);

      out.flush();
      System.exit(64);
    }
  }

  private static Resolver mResolver;

  private static Message query(Name qname, int qtype, int qclass)
  {
    Request request = new Request(qname, qtype, qclass);
    request.setFlag(Flags.RD);

    DNSEvent event = new DNSEvent(request, null);
    // This should synchronously process the request, based on the way the
    // resolver tail was configured.
    mResolver.processRequest(event);
    return event.getResponse().getSMessage().getMessage();
  }

  public static void main(String[] args)
  {

    CLIState state = new CLIState();

    try
    {
      state.parseCommandLine(args);

      NameserverClient nsClient = new NameserverClient();
      // We need a local reference to the stub resolver for the search path.
      StubResolver stubRes = new StubResolver(nsClient);

      if (state.useIterativeResolver)
      {
        mResolver = new ValidatingResolver(new IterativeResolver(nsClient, null));

        // Separately initialize the stubresolver...
        stubRes.init(state.config);
      }
      else
      {
        mResolver = new ValidatingResolver(stubRes);
      }

      // Initialize the resolver chain.
      mResolver.init(state.config);

      // use it:

      Message respMessage;

      if (state.qname.isAbsolute())
      {
        respMessage = query(state.qname, state.qtype, state.qclass);
      }
      else
      {
        List searchList = stubRes.getSearchPath();
        Iterator searchIter = searchList.iterator();

        Name qname = Name.concatenate(state.qname, Name.root);

        respMessage = query(qname, state.qtype, state.qclass);

        while (respMessage.getRcode() == Rcode.NXDOMAIN && searchIter != null
            && searchIter.hasNext())
        {
          Name suffix = (Name) searchIter.next();
          qname = Name.concatenate(state.qname, suffix);

          respMessage = query(qname, state.qtype, state.qclass);
        }
      }

      System.out.println("RESPONSE:");
      System.out.println(respMessage.toString());

    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }
}